package solutions.bitbadger.documents

interface Comparison<T> {

    val op: Op

    val isNumeric: Boolean

    val value: T
}

/**
 * A single-value comparison against a field in a JSON document
 *
 * @property op The operation for the field comparison
 * @property value The value against which the comparison will be made
 */
class SingleComparison<T>(override val op: Op, override val value: T) : Comparison<T> {

    /** Is the value for this comparison a numeric value? */
    override val isNumeric: Boolean
        get() = value.let { it is Byte || it is Short || it is Int || it is Long }

    override fun toString() =
        "$op $value"
}

/**
 * A range comparison against a field in a JSON document
 */
class BetweenComparison<T>(override val op: Op = Op.BETWEEN, override val value: Pair<T, T>) : Comparison<Pair<T, T>> {

    override val isNumeric: Boolean
        get() = value.first.let { it is Byte || it is Short || it is Int || it is Long }
}

/**
 * A check within a collection of values
 */
class InComparison<T>(override val op: Op = Op.IN, override val value: Collection<T>) : Comparison<Collection<T>> {

    override val isNumeric: Boolean
        get() = !value.isEmpty() && value.elementAt(0).let { it is Byte || it is Short || it is Int || it is Long }
}

/**
 * A check within a collection of values
 */
class InArrayComparison<T>(override val op: Op = Op.IN_ARRAY, override val value: Pair<String, Collection<T>>) : Comparison<Pair<String, Collection<T>>> {

    override val isNumeric: Boolean
        get() = !value.second.isEmpty() && value.second.elementAt(0)
            .let { it is Byte || it is Short || it is Int || it is Long }
}
